/*
popclk Version 1.1.0 - World Population Clock
Copyright (C) 2001-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <ncurses.h>
#include <time.h>
#include <errno.h>
#include <sys/times.h>

#define NOTIME 0
#define RELATIVE 1
#define ASCII 2
#define JULIAN 3

int runflg,resetsw,timesw,prevtime;
int helpsw,returnsw,aboutsw;
int exposesw;
int ratesw,titlesw;
time_t frstsecs,totsecs;
double totpop;
double dblanndlta;
double dbltps,dbltpy;
double dayticks,dbltkdlta;
time_t origclk;
int cap_loc;
char caption[128];

void putstx(pgm)
char *pgm;
   {
   fprintf(stderr,"Usage: %s <annual_change> [title]\n",
      pgm);
   fprintf(stderr,"Where annual_change is the "
      "annual increase or decrease\n");
   fprintf(stderr,"from -999999999999999 to "
      "999999999999999\n");
   fprintf(stderr,"Title is optional "
      "and can be more than one word\n");
   fprintf(stderr,"Exit by pressing escape\n");
   exit(1);
   } /* putstx */

void initnc()
   {
   initscr();
   cbreak();
   noecho();
   nonl();
   intrflush(stdscr,FALSE);
   keypad(stdscr,TRUE);
   nodelay(stdscr,TRUE);
   curs_set(0);
   clear();
   refresh();
   } /* initnc */

void escape()
   {
   int ch;
   ch = getch();
   if (ch == 0x1b || ch == 'q')
      {
      runflg = 0;
      }
   else if (ch == 't')
      {
      prevtime = timesw++;
      if (timesw > JULIAN) timesw = NOTIME;
      } /* if toggle time mode */
   else if (ch == 'r')
      {
      resetsw = 1;
      } /* if reset count */
   else if (ch == 'h' || ch == ' ')
      {
      helpsw ^= 1;
      } /* if help */
   else if (ch == 'a')
      {
      aboutsw ^= 1;
      } /* if help */
   else if (ch == '\r')
      {
      returnsw = 1;
      } /* if help */
   else if (ch == 'n')
      {
      ratesw = 1;
      } /* if rate change */
   else if (ch == 'c')
      {
      titlesw = 1;
      } /* if title change */
   } /* escape */

void slp()
   {
   int rslt;
   struct timespec request;
   struct timespec remain;
   request.tv_sec  = (long int) 0;
   request.tv_nsec = (long int) 1;
   rslt = nanosleep(&request,&remain);
   if (rslt < 0)
      {
      fprintf(stderr,"Nanosleep error %d\n",
         errno);
      perror("Nanosleep error");
      exit(1);
      } /* nanosleep error */
   } /* slp using nanosleep */

int readkb()
   {
   int ch;
   while ((ch = getch()) < 0) slp();
   if (ch == 0x0107) return(8);
   return(ch);
   } /* readkb */

void chg_title()
   {
   int ch;
   char *p,*q;
   char str[128];
   char blnk[128];
   char heading[128];
   clear();
   p = blnk;
   q = blnk + 126;
   while (p < q) *p++ = ' ';
   *p     = '\0';
   p = str;
   q = str + 30;
   *p     = '_';
   *(p+1) = '\0';
   strcpy(heading,"Enter Caption:");
   move(1,0);
   addstr(heading);
   ch = 0;
   while (p < q && ch != '\n' && ch != 0x1b)
      {
      move(3,0);
      clrtoeol();
      addstr(str);
      refresh();
      ch = readkb();
      if (ch == 0x1b)
	 {
	 p = str;
	 *p = '\0';
	 break;
	 } /* if cancel */
      if (ch == '\n') break;
      if (ch == '\r') break;
      if (ch == 0x08)
	 {
	 if (p > str)
	    {
   	    p--;
	    *p = '_';
	    *(p+1) = ' ';
	    *(p+2) = ' ';
	    *(p+3) = '\0';
	    } /* if not at beginning */
	 continue;
	 } /* if back space */
      if (ch >= ' ' && ch <= '~')
	 {
         *p     = ch;
         *(p+1) = '_';
         *(p+2) = '\0';
         p++;
	 } /* if valid ASCII */
      else return;
      } /* for each char */

   *p = '\0';  /* override cursor */
   resetsw = titlesw = ratesw = returnsw = helpsw = 0;
   clear();
   if (!strlen(str) || p <= str)
      {
      mvaddstr(0,0,caption);
      return;
      } /* if no change */
   strcpy(caption,str);
   cap_loc = 40 -
      (((strlen(caption) >> 1) + 1) * 8);
   mvaddstr(0,0,caption);
   refresh();
   } /* chg_title */

void chg_rate()
   {
   int ch;
   struct tms t;
   double sign;
   double prevdlta;
   char *p,*q;
   char str[128];
   char blnk[128];
   char heading[128];
   char msg[128];
   clear();
   p = blnk;
   q = blnk + 126;
   while (p < q) *p++ = ' ';
   *p  = '\0';
   p   = str;
   q   = str + 20;
   prevdlta = dblanndlta;
   sign = 1;
   dblanndlta = 0.0;
   *p     = '_';
   *(p+1) = '\0';
   move(1,0);
   strcpy(heading,"Enter Annual Rate:");
   addstr(heading);
   move(3,0);
   addstr(str);
   ch = 0;
   while (p < q && ch != '\n' && ch != 0x1b)
      {
      move(3,0);
      clrtoeol();
      addstr(str);
      refresh();
      ch = readkb();

      if (ch == 0x1b)
	 {
	 dblanndlta = prevdlta;
	 break;
	 } /* if cancel */
      if (ch == '\n') break;
      if (ch == '\r') break;
      if (ch == 0x08)
	 {
	 if (p > str)
	    {
   	    p--;
	    *p = '_';
	    *(p+1) = ' ';
	    *(p+2) = ' ';
	    *(p+3) = '\0';
	    dblanndlta = floor(dblanndlta / 10.0);
	    } /* if not at beginning */
	 else sign = 1;
	 continue;
	 } /* if back space */
      if (ch == '-' && p <= str)
	 {
	 sign   = -1;
	 } /* if sign */
      else if (isdigit(ch))
	 dblanndlta = (dblanndlta * 10.0)
	 + (double) ch - '0';
      else
	 {
	 dblanndlta = prevdlta;
	 return;
	 } /* invalid entry */
      *p     = ch;
      *(p+1) = '_';
      *(p+2) = '\0';
      p++;
      if (dblanndlta >= 1.0e15) return;
      } /* for each digit */

   if (p <= str || dblanndlta == 0.0)
      {
      dblanndlta = prevdlta;
      } /* if no input */

   if (dblanndlta < 1.0e7)
      {
      sprintf(msg,"Warning: low annual rate %.0f",
	 dblanndlta);
      move(5,0);
      addstr(msg);
      sprintf(msg,"Re-enter annual rate?");
      move(6,0);
      addstr(msg);
      sprintf(msg,"Press Y or N");
      move(7,0);
      addstr(msg);
      refresh();
      qiflush();
      ch = readkb();
      if (ch == 'y' || ch == 'Y')
	 {
	 chg_rate();
	 return;
	 } /* if recursion */
      } /* warning */
   if (dblanndlta != prevdlta)
      {
      dblanndlta *= sign;
      dbltkdlta = dblanndlta / dbltpy;
      origclk = times(&t);
      time(&frstsecs);
      totsecs = 0;
      totpop = 0.0;
      } /* if rate changed */

   resetsw = ratesw = titlesw = returnsw = helpsw = 0;
   clear();
   move(0,0);
   addstr(caption);
   refresh();
   } /* chg_rate */


void helpmenu()
   {
   int prevtimesw;
   int prevaboutsw;
   int prevratesw;
   int prevtitlesw;
   char str[128];
   prevtimesw = timesw;
   prevaboutsw = aboutsw;
   prevratesw = ratesw;
   prevtitlesw = titlesw;
   clear();
   while (runflg && !returnsw && helpsw
      && timesw == prevtimesw
      && aboutsw == prevaboutsw
      && ratesw == prevratesw
      && titlesw == prevtitlesw
      && !resetsw)
      {
      strcpy(str,"t) Change time mode");
      move(0,0);
      addstr(str);
      strcpy(str,"r) Reset population count");
      move(1,0);
      addstr(str);
      strcpy(str,"a) About popclk");
      move(2,0);
      addstr(str);
      strcpy(str,"n) Change Annual Rate");
      move(3,0);
      addstr(str);
      strcpy(str,"c) Change Title");
      move(4,0);
      addstr(str);
      strcpy(str,"Return) Return");
      move(5,0);
      addstr(str);
      while (runflg && !returnsw && helpsw
         && timesw == prevtimesw
         && aboutsw == prevaboutsw
         && ratesw == prevratesw
         && titlesw == prevtitlesw
         && !resetsw)
         {
         escape();
         slp();
         } /* wait loop */
      } /* help loop */
   if (ratesw) chg_rate();
   if (titlesw) chg_title();
   clear();
   ratesw = titlesw = returnsw = helpsw = 0;
   mvaddstr(0,0,caption);
   refresh();
   } /* helpmenu */

void disclaim()
   {
   int ch;
   clear();
   mvaddstr(1,0,"popclk.c version 1.1.0");
   mvaddstr(2,0,"Copyright (C) 2001-2010  ");
   addstr("dondalah721@yahoo.com (Dondalah)");
   mvaddstr(3,0,"popclk.c comes with ABSOLUTELY ");
   addstr("NO WARRANTY");
   mvaddstr(4,0,"For details, read the GNU ");
   addstr("Public License at");
   mvaddstr(5,0,"http://www.gnu.org/copyleft/gpl.html");
   mvaddstr(6,0,"Press any key to continue.");
   curs_set(0);
   move(0,0);
   refresh();
   while ((ch = getch()) == -1);
   clear();
   move(0,0);
   addstr(caption);
   refresh();
   helpsw = aboutsw = 0;
   } /* disclaim */

void shwtps(ticks)
unsigned int ticks;
   {
   int ch;
   char str[32];
   sprintf(str,"Ticks per second %u", ticks);
   addstr(str);
   move(2,0);
   addstr("Press enter, escape to exit");
   refresh();
   ch = -1;
   while (ch < 0)
      {
      ch = getch();
      if (ch == 0x1b)
         {
         endwin();
         exit(0);
         } /* if escape */
      slp();
      } /* wait for key */
   clear();
   refresh();
   } /* shwtps */

int gettps()
   {
   unsigned int diff;
   time_t frstsec,currsec;
   clock_t prev,clk;
   struct tms t;
   clear();
   mvaddstr(1,0,"h = Help");
   mvaddstr(2,0,"t = Time mode");
   mvaddstr(3,0,"r = Reset count");
   mvaddstr(4,0,"n = Change rate");
   mvaddstr(5,0,"r = Change caption");
   mvaddstr(6,0,"a = About popclk");
   mvaddstr(7,0,"<return> = return/quit");
   refresh();
   /* get first date & time in seconds */
   time(&frstsec);
   currsec = frstsec;
   while (currsec == frstsec)
      {
      /* get first date & time in seconds */
      time(&currsec);
      slp();
      } /* loop until second starts */
   /* get first clock ticks since boot */
   prev = times(&t);
   frstsec = currsec;
   while (currsec == frstsec)
      {
      /* get current date & time in seconds */
      time(&currsec);
      slp();
      } /* loop for 1 second */
   /* get clock ticks since boot */
   clk = times(&t);
   if (clk < prev)
      {
      fprintf(stderr,"Negative ticks, rerun job\n");
      exit(0);
      }
   diff = clk - prev;
   sleep(1);
   clear();
   refresh();
   return(diff);
   } /* gettps */

void prttime(void)
   {
   int hrs,mins,secs,len;
   double dblhour,dbljul;
   double greg2jd();
   char *asc;
   char str[64];
   time_t now,relsecs;
   struct tm *lcltm;
   if (timesw == RELATIVE)
      {
      time(&now);
      relsecs = now - frstsecs;
      if (relsecs > totsecs)
         {
         totsecs = relsecs;
         hrs = relsecs / 3600;
         mins = (relsecs % 3600) / 60;
         secs = relsecs % 60;
         sprintf(str,"      %02d:%02d:%02d",
            hrs, mins, secs);
         move(5,0);
         clrtoeol();
         addstr(str);
         move(0,0);
         refresh();
         } /* if second elapsed */
      } /* relative time */
   else if (timesw == ASCII)
      {
      time(&now);
      relsecs = now - frstsecs;
      if (relsecs > totsecs)
         {
         totsecs = relsecs;
         lcltm = (struct tm *) localtime(&now);
         asc = (char *) asctime(lcltm);
         len = strlen(asc);
         asc[len-1] = '\0';
         move(5,0);
         clrtoeol();
         addstr(asc);
         move(0,0);
         refresh();
         } /* if second elapsed */
      } /* ascii */
   else if (timesw == JULIAN)
      {
      time(&now);
      relsecs = now - frstsecs;
      if (relsecs > totsecs)
         {
         totsecs = relsecs;
         lcltm = (struct tm *) localtime(&now);
         dblhour = (double) lcltm->tm_hour
            + (double) lcltm->tm_min / 60.0
            + (double) lcltm->tm_sec / 3600.0;
         dbljul = greg2jd(lcltm->tm_year+1900,
            lcltm->tm_mon+1,
            lcltm->tm_mday,
            dblhour);
         sprintf(str,"    %f", dbljul);
         move(5,0);
         clrtoeol();
         addstr(str);
         move(0,0);
         refresh();
         } /* if second elapsed */
      } /* ascii */
   else
      { /* no time */
      if (timesw != prevtime)
         {
         prevtime = timesw;
         move(5,0);
         clrtoeol();
         } /* if change in time status */
      } /* no time */
   } /* prttime */

void count()
   {
   unsigned int diff;
   double tgtpop,currpop;
   time_t clk;
   struct tms t;
   char str[64];
   currpop = totpop;
   if (fabs(totpop) < 1.0)
      {
      sprintf(str,"%10.0f", totpop);
      move(2,2);
      clrtoeol();
      addstr(str);
      } /* if zero total */
   if (dbltkdlta > 0.0)
      tgtpop = totpop + 1.0;
   else
      tgtpop = totpop - 1.0;
   while (runflg && !returnsw
      && !resetsw && !ratesw)
      {
      if (dbltkdlta > 0.0
         && currpop > tgtpop) break;
      else if (dbltkdlta <= 0.0
         && currpop < tgtpop) break;
      clk = times(&t);
      if (clk < origclk)
         diff = clk + dayticks - origclk;
      else
         diff = clk - origclk;
      /* recalculate the total */
      /* population from scratch */
      currpop = dbltkdlta * diff;
      slp();
      prttime();
      escape();
      if (helpsw)
	 {
	 helpmenu();
	 break;
	 } /* if help menu */
      if (aboutsw)
	 {
	 disclaim();
	 break;
	 } /* if aboutsw */
      if (titlesw)
	 {
	 chg_title();
	 break;
	 } /* if change title */
      } /* wait for birth */
   if (resetsw)
      {
      resetsw = 0;
      currpop = 0.0;
      origclk = clk;
      totsecs = 0;
      time(&frstsecs);
      } /* if reset */
   totpop = currpop;
   sprintf(str,"%10.0f", totpop);
   move(2,2);
   clrtoeol();
   addstr(str);
   prttime();
   move(0,0);
   refresh();
   } /* count */

void runclk()
   {
   while (runflg && !returnsw)
      {
      count();
      escape();
      if (!runflg) return;
      slp();
      if (ratesw) chg_rate();
      if (titlesw) chg_title();
      if (helpsw) helpmenu();
      if (aboutsw) disclaim();
      } /* display loop */
   } /* runclk */

int main(argc,argv)
int argc;
char **argv;
   {
   unsigned int tps;
   double dblspy,sign;
   double dbltps;
   double ten,unit;
   char *p;
   struct tms t;
   if (argc < 2) putstx(*argv);
   p = *(argv+1);
   if (*p == '-')
      {
      sign = -1.0;
      p++;
      }
   else sign = 1.0;
   dblanndlta = 0.0;
   ten = 10.0;
   while (*p != '\0')
      {
      if (*p < '0' || *p > '9')
         putstx(*argv);
      unit = (double) *p - '0';
      dblanndlta = (dblanndlta * ten) + unit;
      p++;
      }
   if (dblanndlta >= 1.0e15)
      putstx(*argv);
   if (dblanndlta < 1.0e7)
      {
      fprintf(stderr,"Warning: low annual rate %.0f\n",
	 dblanndlta);
      fprintf(stderr,"Press <Return> to continue.\n");
      getchar();
      }
   /* annual change in population */
   dblanndlta *= sign;
   /* seconds per year */
   dblspy = 3600.0 * 24.0 * 365.25;
   initnc();
   /* ticks per second */
   tps = gettps();
   if (tps < 1 || tps > 999999)
      {
      fprintf(stderr,"Invalid tps %d\n", tps);
      exit(1);
      } /* if tps error */
   dbltps = (double) tps;
   /* ticks per year */
   dbltpy = dblspy * dbltps;
   /* population change per tick */
   dbltkdlta = dblanndlta / dbltpy;
   /* ticks per day */
   dayticks = 3600 * 24 * tps;
   /* initialize the curses screen */
   strcpy(caption,"World Population Clock");
   if (argc > 2)
      {
      int i;
      strcpy(caption,*(argv+2));
      i = 3;
      while (i < argc)
	 {
	 strcat(caption," ");
	 strcat(caption,*(argv + (i++)));
	 } /* for each word in title */
      } /* if title on command line */
   mvaddstr(0,0,caption);
   /* start counting now */
   timesw = prevtime = RELATIVE;
   origclk = times(&t);
   time(&frstsecs);
   resetsw = totsecs =
      helpsw = aboutsw =
      ratesw = titlesw = 0;
   totpop = 0.0;
   runflg = 1;
   runclk();
   endwin();
   printf("Ticks per second: %d\n", tps);
   return(0);
   } /* main */
